#include <Arduino.h>
#include "ICMD.h"

  static int ICMD::instances;
  static  ICMD  * ICMD::instance[MAXINSTANCES];
  
//
// Class to control a ICMD counter IC from IC-Haus
//   using the SIKTEC-SPI-library for Soft and Hardware SPI
//
// Note: not all modes supported
//
// Optional Defines:
//    OVERFLOW_HANDLER
//      - Connect NWARN of IC-MD to an interrupt-capable warnpin in Arduino
//      - Use ICMD Constructor with warnpin
//  
//    CNTCFG: 
//      SINGLE_TTL_24BIT   
//      SINGLE_TTL_48BIT   
//      SINGLE_TTL_32BIT  
//      TWO_TTL_24BIT      
//      TWO_TTL_16BIT      
//      THREE_TTL_16BIT    
//
//    HW_SPI - Default is Software SPI
//        
//    RESET_CNT_ON_Z
//
ICMD::init(int slave_cs_pin,int slave_sck_pin, int slave_miso_pin,int slave_mosi_pin)
  {
    byte buf[16];
#ifndef HW_SPI 
  this->spi = new SIKTEC_SPI(slave_cs_pin, slave_sck_pin, slave_miso_pin,slave_mosi_pin, 1000000,MSBFIRST);
#else
  this->spi = new SIKTEC_SPI(slave_cs_pin, 1000000,MSBFIRST);
#endif


 
  ICMD::instance[ICMD::instances] = this;
  ICMD::instances++;
  
  for(int i=0;i<3;i++)
  {
    this->turns[i]=0;
    this->dir[i]=1;
    this->oldCount[i]=0;
  }
  
  pinMode(SS, OUTPUT);
  pinMode(slave_miso_pin, INPUT_PULLUP);
  pinMode(slave_mosi_pin, OUTPUT);
  pinMode(slave_sck_pin, OUTPUT);

  this->spi->begin(); //??
  this->spi->beginTransaction ();
  buf[0]=0x00;      // Config register 0

  buf[1]=CNTCFG; // Counter Configuration

// Encoders with ZERO-Signal?
 #ifdef RESET_CNT_ON_Z  
  buf[1] += 0x60; // TTL, CNT0+1 cleared by Z0 active when A=B=0, funktioniert momentan nicht
 #endif
 
  this->spi->write(buf,2);
  this->spi->endTransaction();
   
  this->spi->beginTransaction ();
  this->spi->read(buf,1,0x80 | 0x00 );
  Serial.print( buf[1],BIN); 
  Serial.print(" ");
  
  this->spi->endTransaction();
  this->spi->beginTransaction ();
  buf[0]=0x01;       // Config register 0
  buf[1]=0x80+0x18;  // INVZ1, CCW positive for Counter 0 and 1

  this->spi->write(buf,2);
  this->spi->endTransaction();
  /*
  spi.beginTransaction ();
  buf[0]=0x30;      // Instruction Byte
  buf[1]=0x09;      // Reset Counter 0 , Enable Zero Codification, Load TP??
  spi.write(buf,2);
  spi.endTransaction();
*/

// Reset counters
  this->spi->beginTransaction ();
  buf[0]=0x30;      // Instruction Byte
  buf[1]=0x07;      // Restet all counters
  this->spi->write(buf,2);
  this->spi->endTransaction();



  Serial.begin(250000); //initialize serial comunication
  Serial.print("IC MD RX ");
 

// Read the manufacturer ID to see that Chip, Wiring and SPI interface are ok
 this->readManufacturerID();
}

ICMD::ICMD(int slave_cs_pin,int slave_sck_pin, int slave_miso_pin,int slave_mosi_pin)
{
   init(slave_cs_pin,slave_sck_pin,slave_miso_pin,slave_mosi_pin);
}


ICMD::ICMD(int slave_cs_pin,int slave_sck_pin, int slave_miso_pin,int slave_mosi_pin,int wp) 
  {
    byte buf[16];
    init(slave_cs_pin,slave_sck_pin,slave_miso_pin,slave_mosi_pin);
    this->warnpin=wp; 
     
    Serial.println("OVERFLOW_HANDLER");
    pinMode(this->warnpin, INPUT);

    int checkPin = digitalPinToInterrupt(this->warnpin);

    if (checkPin == -1) {
      Serial.print("Pin ");
      Serial.print(this->warnpin);
      Serial.println(" not a valid interrupt pin!");
   }
    else
      attachInterrupt(checkPin, this->overflowHandler, FALLING); 


  this->spi->beginTransaction ();
  buf[0]=0x02;      // MASK Register
  buf[1]=0x08;      // zero and overflow of counter 0,1,2 to NWARN-Pin
  this->spi->write(buf,2);
  
  this->spi->endTransaction();
  readSTATUS(0,buf);
  readSTATUS(2,buf);
  readSTATUS(1,buf);
  Serial.println("OVERFLOW_HANDLER");


  }
  

//
// experimental overflow-handling for instance 0 !! 
//
void ICMD::overflowHandler(){
   byte buf[3];
   byte response[10];
   /*
   ICMD::instance[0]->spi->beginTransaction ();
   ICMD::instance[0]->spi->read(response,3,0x80 | 0x0A ); 
   ICMD::instance[0]->spi->endTransaction();
*/

   
   Serial.println("**");
   for(int i=0;i<3;i++)
   {
     ICMD::instance[0]->readSTATUS(i,buf); 
     Serial.print(i); Serial.print(" "); Serial.println(buf[1],HEX);
     if (buf[1] & 0x40) // Overflow in Bit 6
     {

        ICMD::instance[0]->readData();
        
   
        long count = ICMD::instance[0]->getCounter( ICMD::instance[0]->response,i);
        Serial.print("OVERFLOW Counter ");
        Serial.println(i);
        Serial.print("oldCount=");
        Serial.println(ICMD::instance[0]->oldCount[i]);
        if(ICMD::instance[0]->oldCount[i]>0 )
        {
          ICMD::instance[0]->dir[i]=1;
          ICMD::instance[0]->turns[i]++;
        }
        else
        {
          ICMD::instance[0]->dir[i]= -1;
          ICMD::instance[0]->turns[i]--;
        }
     }
   }
   Serial.println();
}


//
// conversion routines to get counter values fromn byteArray
// 

long ICMD::Bit32ToInt32(byte *byteArray)
{
  /*
   Serial.print(byteArray[0],HEX);
   Serial.print(byteArray[1],HEX);
   Serial.print(byteArray[2],HEX);
   Serial.print(byteArray[3],HEX);
   Serial.print(" ");
   */
          //https://stackoverflow.com/questions/5678049/difference-between-uint-and-unsigned-int
   long result = ( 
        (((ulong)0xFF & byteArray[0]) << 24) | 
         ((ulong)(0xFF & byteArray[1]) << 16) | 
         ((ulong) (0xFF & byteArray[2]) << 8) |  
          (ulong) (0xFF & byteArray[3]) 
         ); 
   if ((result & (ulong)0x0080000000) > 0)
   { 
       result = (long)((ulong)result|(ulong)0xFF000000); 
   }
   else
   { 
       result = (long)((ulong)result & (ulong)0x00FFFFFFFF); 
   } 
   return result;
}

long ICMD::Bit24ToInt32(byte *byteArray)
{
  /*
   Serial.print(byteArray[0],HEX);
   Serial.print(byteArray[1],HEX);
   Serial.print(byteArray[2],HEX);
   Serial.print(" ");
   */
          //https://stackoverflow.com/questions/5678049/difference-between-uint-and-unsigned-int
   long result = ( 
      ((long) (0xFF & byteArray[0]) << 16) | 
      ((long) (0xFF & byteArray[1]) << 8) |  
       (long) (0xFF & byteArray[2]) 
    ); 
   if ((result & (long) 0x00800000) > 0)
   { 
     result = (long)((ulong)result|(ulong)0xFF000000); 
   }
   else
   { 
     result = (long)((ulong)result & (ulong)0x00FFFFFF); 
   } 
  
   return result;
}

int ICMD::Bit16ToInt32(byte *byteArray)
{
          //https://stackoverflow.com/questions/5678049/difference-between-uint-and-unsigned-int
   int result = ( 
       ((0xFF & byteArray[0]) << 8) |  
        (0xFF & byteArray[1]) 
   ); 
   if ((result & 0x008000) > 0)
   { 
     result = (int)((uint)result|(uint)0xFF0000); 
   }
   else
   { 
     result = (int)((uint)result & (uint)0x00FFFF); 
   } 
   return result;
}


/*     
long ICMD::readUPD()
{
   byte buf[10];
 
  digitalWrite(SS, LOW);
  SPI.transfer( 0x80 + 0x0A);
   for(int i=0;i<3;i++)
   {
      buf[i]=SPI.transfer( 0);
   }

  
   Serial.print(" UPD: ");
    for(int i=0;i<3;i++)
   {
    Serial.print(buf[i]); Serial.print(" ");
   }
   Serial.print(" ");
   digitalWrite(SS, HIGH);
}
*/

void static ICMD::readSTATUS( int i, byte * buf)
{
  this->spi->beginTransaction();
  this->spi->read( buf,2, 0x80 + 0x48+i); // read Status Register #i, 0x80 for read
  
    Serial.print("Status:"); Serial.print(buf[1],HEX); Serial.print(" "); 
   this->spi->endTransaction();
}

byte * ICMD::readData()
{
  byte buf[10];
  for(int i=0;i<DATALEN;i++)
  {
       this->responseOld[i] = this->response[i];

   
      }    
  this->spi->beginTransaction ();
  this->spi->read(this->response,DATALEN,0x80 | 0x08 ); 
  this->spi->endTransaction();
  //this->printResponse();
  /*
  for(int i=0;i<DATALEN;i++) {Serial.print(response[i]); Serial.print(" ");}
  Serial.println();
  */
  return this->response;
}

int ICMD::hasChanged()
{
  for(int i=1;i<DATALEN;i++)
    if(this->response[i] != this->responseOld[i]) { return 1;}
  return 0;
}


//
// Read the Ref
//
long ICMD::readREF()
{
    byte buf[10];
 
    this->spi->beginTransaction();
    this->spi->read( buf,3, 0x80 | 0x10); // read ref Register, 0x80 for read
    Serial.print("REF:");
    Serial.print(Bit24ToInt32(buf+1)); Serial.print(" ");
    this->spi->endTransaction();
    return Bit24ToInt32(buf+1);
}

void ICMD::printResponse()
  {
    for(int i=0;i<DATALEN;i++)
    {
      Serial.print(this->response[i]); Serial.print(" ");
    }
    Serial.println();
  
  }
  
//
// read the TP
//
long ICMD::readTP()
{
   byte buf[10];
 
   this->spi->beginTransaction();  // read TP1 Register, 0x80 for read
   this->spi->read( buf,3, 0x80 + 0x0C );
   Serial.print("TP1:");
   Serial.print(Bit24ToInt32(buf+1)); Serial.print(" ");
   this->spi->endTransaction();

   this->spi->beginTransaction();
   this->spi->read( buf,3, 0x80 + 0x0E);  // read TP2 Register, 0x80 for read
   Serial.print("TP2:");
   Serial.print(Bit24ToInt32(buf+1)); Serial.print(" ");
   this->spi->endTransaction();
   
   return -1;
}

//
// print Counter values:
//
void ICMD::printCounter(byte * response)
{
  /*
  for(int i=0;i<DATALEN;i++)
    {Serial.print(response[i]);
    Serial.print(" ");
    }
    Serial.println();
    return;
    */
  #if CNTCFG== TWO_TTL_16BIT
      // 2 16-Bit TTL-Counter
      Serial.print("2x16 CNT0:");
   Serial.print(Bit16ToInt32(response+3));
   Serial.print(" CNT1:");
   Serial.print(Bit16ToInt32(response+1));
   // Serial.print(Bit16ToInt32(response));
   //Serial.print(" CNT1:");
   //Serial.print(Bit16ToInt32(response+2));
  #endif
  #if CNTCFG==  SINGLE_TTL_24BIT
    // 1 24-Bit TTL-Counter
     Serial.print("1x24 CNT0:");
   Serial.print(Bit24ToInt32(response+1)); 
  #endif
  
  #if CNTCFG==  SINGLE_TTL_32BIT
    // 1 32-Bit TTL-Counter
   Serial.print("1x32 CNT0:");
   Serial.print(Bit32ToInt32(response+1)); 
  #endif
  
  #if CNTCFG==  TWO_TTL_24BIT
    // 2 24-Bit TTL-Counter
     Serial.print("2x24 CNT0:");
   Serial.print(Bit24ToInt32(response+4)); 
   Serial.print(" ");
    Serial.print("CNT1:");
   Serial.print( Bit24ToInt32(response+1) ); 
  #endif
  
   Serial.print(" | ");
}
void ICMD::printCounter()
{
  this->printCounter(this->response);
}
//
// Get Counter Values from a response
// /!\ WORK IN PROGRESS - only for 2 counters
//
long ICMD::getCounter(byte * response, int ic)
{
  long count=0;
  #if CNTCFG== TWO_TTL_16BIT
      // 2 16-Bit TTL-Counter
      if(ic==1)
  count= Bit16ToInt32(response+1);
else
  count= (Bit16ToInt32(response+3));
  #endif
  #if CNTCFG==  SINGLE_TTL_24BIT
    // 1 24-Bit TTL-Counter
   count= Bit24ToInt32(response+1); 
  #endif
  
  #if CNTCFG==  SINGLE_TTL_32BIT
    // 1 32-Bit TTL-Counter
   count = Bit32ToInt32(response+1); 
  #endif
  
  #if CNTCFG==  TWO_TTL_24BIT
    // 2 24-Bit TTL-Counter
         if(ic==1)
   count= Bit24ToInt32(response+1);
  else
   count=Bit24ToInt32(response+4) ; 
  #endif
  
  

  if(count != this->oldCount)
  {
      if (count > this->oldCount[ic]) 
        this->dir[ic]=1;
        else
        this->dir[ic]= -1;    
  }
  this->oldCount[ic]=count;
    
  //    Serial.print(count + (long)turns *  (long)((long)1<<16)); Serial.print(" : "); Serial.print(turns);   
  return count;     
}

long ICMD::getCounter( int ic)
{
  return this->getCounter(this->response,ic);
}

//
// reset counter c
//
void ICMD::resetCounter(int c)
{ 
   byte buf[3];
   int bits = 1<< c; 

   this->spi->beginTransaction ();
   buf[0]=0x30;
   buf[1]=bits; // only CNT0
  // Serial.print("reset ");
  // Serial.println(bits);
  this->spi->write(buf,2);
  this->spi->endTransaction ();
  turns[c]=0; 
}


void ICMD::readManufacturerID()
{
  byte buf[10];
  this->spi->begin();
  this->spi->beginTransaction();
  this->spi->read( buf,8,0x80 + 0x78);
  this->spi->endTransaction();
  for(int i=0;i<8;i++)
  {
    Serial.print(buf[i+1]); Serial.print(" ");
  }
  Serial.println("");
}
