#ifndef ICMD_h
#define OVERFLOW_HANDLER

#include <SPI.h>
#include <SIKTEC_SPI.h>

typedef  unsigned int     uint; 
typedef  unsigned long    ulong;

//
// Default Pins for Software-SPI for Arduino MEGA
//
#ifndef SLAVE_CS_PIN
#define SLAVE_CS_PIN 17
#endif

#ifndef SLAVE_SCK_PIN
#define SLAVE_SCK_PIN 52 
#endif

#ifndef SLAVE_MISO_PIN
#define SLAVE_MISO_PIN 50
#endif

#ifndef SLAVE_MOSI_PIN
#define SLAVE_MOSI_PIN 51
#endif


//
// Counter Configurations
//
#define SINGLE_TTL_24BIT   0x00
#define SINGLE_TTL_48BIT   0x02
#define SINGLE_TTL_32BIT   0x04
#define TWO_TTL_24BIT      0x01
#define TWO_TTL_16BIT      0x06
#define THREE_TTL_16BIT    0x07


//
// Default Counter Configuration 
//
#ifndef CNTCFG
  //CNTCFG undefined!
  //#define CNTCFG TWO_TTL_24BIT
  #define CNTCFG TWO_TTL_16BIT
#endif

#if CNTCFG == SINGLE_TTL_48BIT
#define DATALEN 9
#endif

#if CNTCFG == SINGLE_TTL_32BIT
#define DATALEN 5
#endif

#if CNTCFG == THREE_TTL_16BIT
#define DATALEN 5
#endif

#if CNTCFG == SINGLE_TTL_24BIT
#define DATALEN 5
#endif

#if CNTCFG == TWO_TTL_16BIT
#define DATALEN  5

#endif

#if CNTCFG == TWO_TTL_24BIT
#define DATALEN 7
//9
#endif


#define ICMD_h

#define MAXINSTANCES 4

class ICMD
{
  public:
  static int instances;
  static  ICMD * instance[MAXINSTANCES];

  volatile byte response[DATALEN];
  volatile byte responseOld[DATALEN];
  
  //Create a SIKTEC_SPI spi object:
  SIKTEC_SPI * spi;
  int warnpin;
  long oldCount[3];
  int turns[3];
  int dir[3];
  
  ICMD(int slave_cs_pin,int slave_sck_pin, int slave_miso_pin,int slave_mosi_pin);
  //
  // experimental with Overflow-Handler
  //
  ICMD(int slave_cs_pin,int slave_sck_pin, int slave_miso_pin,int slave_mosi_pin, int warnpin);

  long Bit32ToInt32(byte *byteArray);
  long Bit24ToInt32(byte *byteArray);
  int Bit16ToInt32(byte *byteArray);
  void readSTATUS( int i, byte * buf);
  byte *  readData();
  int hasChanged();
  
  long readREF();
  long readTP();
  void printResponse();
  void printCounter(byte * response);
  void printCounter();
  long getCounter(byte * response, int ic);
  long getCounter(int ic);
  
  void resetCounter(int bits);
  void readManufacturerID();
  void static overflowHandler();
  
private:  
    init(int slave_cs_pin,int slave_sck_pin, int slave_miso_pin,int slave_mosi_pin);
  
};


#endif
