from casadi import *  
import math

#
# Forward und Backward Kinematik  mit Casadi für
# einen Cablebot mit 4 Seilen in einer quaderförmigen Box, 
# der 3 translatorische Freiheitsgrade aufweist
# A= Breite, B=Höhe
#
# Seilrolle 1: (0,0,B)
# Seilrolle 2: (A,0,B)
# Seilrolle 3: (A/2,A,B)
# Seilrolle 4: (A/2,A/2,0)
#
# Der ideale Endeffektor ist punktfürmig.
# Der reale Endeffektor ist ein gleichseitiges Dreieck mit Seitenlänge c
#
# Dimensionen der Box:
A=500
B=500

# Endeffektor (Gleichseitige Dreieck)  (>0 dann nicht mehr eindeutig)
c=30 # Seitenlänge des Endeffektors
m = math.sqrt( c**2 / (2-2*math.cos(120/180*math.pi))) # Dreiecksmitte-Spitze
h = math.sqrt( 3/4*c**2) # Höhe Gleichseiteiges Dreieck
g=h-m # Dreiecksmitte - Mitte einer Seite
d = 5 # Dicke des Endeffektors

# variables (Koordinaten)
x = SX.sym('x',1)
y = SX.sym('y',1)
z = SX.sym('z',1)

# parameters (Seillängen)
l1 = SX.sym('l1',1)
l2 = SX.sym('l2',1)
l3 = SX.sym('l3',1)
l4 = SX.sym('l4',1)
params=[l1,l2,l3,l4]

#equations  
t1=(B-z)**2
t2=(y-g)**2
t3=(x-c/2)**2
t4=(A-x-c/2)**2
t5=(A-y-m)**2
t6=(A/2 -x)**2
t7=(y-A/2)**2

g1 = l1 - sqrt( t1+t2+t3 )
g2 = l2 - sqrt( t1+t2+t4 )
g3 = l3 - sqrt( t1+t5+t6 )
g4 = l4 - sqrt( (z-d)**2 + t7 + t6 )

iparams = [x,y,z]
params  = [l1,l2,l3]
params2  = [l2,l3,l4]

#
# System für Rüwckwärts-Kinematik ist überbestimmt
# => jeweils 2 Gleichungen
#
                  # Variables  Paraemters         Equateions
g_fwd  = Function('gfwd',[vertcat(x,y,z),    vertcat(*params)],     [vertcat(g1,g2,g3)])
g_fwd2  = Function('gfwd2',[vertcat(x,y,z),    vertcat(*params2)],     [vertcat(g2,g3,g4)])
g_bwd  = Function('gbwd',[vertcat(l1,l2,l3), vertcat(*iparams)], [vertcat(g1,g2,g3)])
g_bwd2  = Function('gbwd2',[vertcat(l2,l3,l4), vertcat(*iparams)], [vertcat(g2,g3,g4)])

opts = {}
opts["nlpsol"] = "ipopt"
opts["nlpsol_options"] = {"ipopt.tol":1e-14}
G_fwd=rootfinder("G_fwd", "nlpsol", g_fwd, opts)
G_fwd2=rootfinder("G_fwd2", "nlpsol", g_fwd2, opts)
G_bwd=rootfinder("G_bwd", "nlpsol", g_bwd, opts)
G_bwd2=rootfinder("G_bwd2", "nlpsol", g_bwd2, opts)

# Zielkoordinaten:
p=[400,250,50]
soli=G_bwd([400,400,200],p)
soli2=G_bwd2([400,400,200],p)
sol=G_fwd([210,150,250],soli)
sol2=G_fwd2([210,150,250],soli2)

print("Seillängen:")
print("l1,l2,l3:",soli)
print("l2,l3,l4:",soli2)
print("Koordinaten:")
print(sol)
print(sol2)

