import numpy as np
import math
from casadi import *  
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d.axes3d as p3
import matplotlib.animation as animation


#
# Forward und Backward Kinematik  mit Casadi für
# einen Cablebot mit 4 Seilen in einer quaderförmigen Box, 
# der 3 translatorische Freiheitsgrade aufweist
# A= Breite, B=Höhe
#
# Seilrolle 1: (0,0,B)
# Seilrolle 2: (A,0,B)
# Seilrolle 3: (A/2,A,B)
# Seilrolle 4: (A/2,A/2,0)
#
# Der ideale Endeffektor ist punktfürmig.
# Der reale Endeffektor ist ein gleichseitiges Dreieck mit Seitenlänge c
#
# Dimensionen der Box:
A=500
B=500

# Endeffektor (Gleichseitige Dreieck)  (>0 dann nicht mehr eindeutig)
c=30 # Seitenlänge des Endeffektors
m = math.sqrt( c**2 / (2-2*math.cos(120/180*math.pi))) # Dreiecksmitte-Spitze
h = math.sqrt( 3/4*c**2) # Höhe Gleichseiteiges Dreieck
g=h-m # Dreiecksmitte - Mitte einer Seite
d = 5 # Dicke des Endeffektors

# variables (Koordinaten)
x = SX.sym('x',1)
y = SX.sym('y',1)
z = SX.sym('z',1)

# parameters (Seillängen)
l1 = SX.sym('l1',1)
l2 = SX.sym('l2',1)
l3 = SX.sym('l3',1)
l4 = SX.sym('l4',1)
params=[l1,l2,l3,l4]

#equations  
t1=(B-z)**2
t2=(y-g)**2
t3=(x-c/2)**2
t4=(A-x-c/2)**2
t5=(A-y-m)**2
t6=(A/2 -x)**2
t7=(y-A/2)**2

g1 = l1 - sqrt( t1+t2+t3 )
g2 = l2 - sqrt( t1+t2+t4 )
g3 = l3 - sqrt( t1+t5+t6 )
g4 = l4 - sqrt( (z-d)**2 + t7 + t6 )

iparams = [x,y,z]
params  = [l1,l2,l3]
params2  = [l2,l3,l4]

#
# System für Rüwckwärts-Kinematik ist überbestimmt
# => jeweils 2 Gleichungen
#
                  # Variables  Paraemters         Equateions
g_fwd  = Function('gfwd',[vertcat(x,y,z),    vertcat(*params)],     [vertcat(g1,g2,g3)])
g_fwd2  = Function('gfwd2',[vertcat(x,y,z),    vertcat(*params2)],     [vertcat(g2,g3,g4)])
g_bwd  = Function('gbwd',[vertcat(l1,l2,l3), vertcat(*iparams)], [vertcat(g1,g2,g3)])
g_bwd2  = Function('gbwd2',[vertcat(l2,l3,l4), vertcat(*iparams)], [vertcat(g2,g3,g4)])

opts = {}
opts["nlpsol"] = "ipopt"
opts["nlpsol_options"] = {"ipopt.tol":1e-14}
G_fwd=rootfinder("G_fwd", "nlpsol", g_fwd, opts)
G_fwd2=rootfinder("G_fwd2", "nlpsol", g_fwd2, opts)
G_bwd=rootfinder("G_bwd", "nlpsol", g_bwd, opts)
G_bwd2=rootfinder("G_bwd2", "nlpsol", g_bwd2, opts)


def calcLengths(x,y,z):
  p=[x,y,z]
  soli=G_bwd([400,400,200],p)
  soli2=G_bwd2([400,400,200],p)
  
  return (soli[0],soli[1],soli[2],soli2[2])


def calcPos(l1,l2,l3,l4):
  
  sol=G_fwd([210,150,250],[l1.__float__(),l2.__float__(),l3.__float__()])
  sol2=G_fwd2([210,150,250],[l2.__float__(),l3.__float__(),l4.__float__()])
  print("P=",sol)
  print("P'=",sol2)
  
  return (sol)


def Gen_line(x0,y0,z0,x1,y1,z1):
    """
    Create a line using a random walk algorithm

    length is the number of points for the line.
    dims is the number of dimensions the line has.
    """
    lineData = np.empty((3, 2))
    lineData[0,0] = x0
    lineData[1,0] = y0
    lineData[2,0] = z0
    lineData[0,1] = x1
    lineData[1,1] = y1
    lineData[2,1] = z1
    
    return lineData


PX = [A/2,A/2,B/2]
def update_lines(num, dataLines, lines):
    #PX[2]=PX[2]-1
    for line, data in zip(lines, dataLines):
        data = [ Gen_line(0,0,B,PX[0],PX[1],PX[2]), Gen_line(0,A,B,PX[0],PX[1],PX[2]), Gen_line(A,A/2,B,PX[0],PX[1],PX[2]),Gen_line(A/2,A/2,0,PX[0],PX[1],PX[2])]

        # NOTE: there is no .set_data() for 3 dim data...
        line.set_data(data[0:2, :num])
        line.set_3d_properties(data[2, :num])
    return lines

P0 =  (A/2,A/2,B/2)
P1 =  (0,0,B)
P2 =  (0,A,B)
P3 =  (A,A/2,B)
P4 =  (A/2,A/2,0)


L0 =  calcLengths(P0[0],P0[1],P0[2])
print("L0=",L0)
PX=calcPos(L0[0],L0[1],L0[2],L0[3])
print("PX=",PX)
 

# Attaching 3D axis to the figure
fig = plt.figure()
ax = p3.Axes3D(fig,auto_add_to_figure=False ) 
fig.add_axes(ax) 
# Fifty lines of random 3-D lines
data = [ Gen_line(0,0,B,PX[0],PX[1],PX[2]), Gen_line(0,A,B,PX[0],PX[1],PX[2]), Gen_line(A,A/2,B,PX[0],PX[1],PX[2]),Gen_line(A/2,A/2,0,PX[0],PX[1],PX[2])]

# Creating fifty line objects.
# NOTE: Can't pass empty arrays into 3d version of plot()
lines = [ax.plot(dat[0, 0:5], dat[1, 0:5], dat[2, 0:5])[0] for dat in data]

# Setting the axes properties
ax.set_xlim3d([0.0, 500.0])
ax.set_xlabel('X')

ax.set_ylim3d([0.0, 500.0])
ax.set_ylabel('Y')

ax.set_zlim3d([0.0, 500.0])
ax.set_zlabel('Z')

ax.set_title('3D Test')

# Creating the Animation object
line_ani = animation.FuncAnimation(fig, update_lines, 25, fargs=(data, lines),
                                   interval=50, blit=False)

plt.show()